import base64
import uuid
import tarfile
import importlib
import shutil
import os

from slai.loaders import base_loader

TMP_MODEL_ARTIFACT_PATH_ROOT = "/tmp/artifacts"


class KerasLoader(base_loader.BaseLoader):
    @classmethod
    def load_model(cls, model_metadata, model_data):
        loaded_model = None
        inference_method_name = None

        tensorflow = importlib.import_module("tensorflow")
        model_tmp_location = f"{TMP_MODEL_ARTIFACT_PATH_ROOT}/{uuid.uuid4()}"

        model_artifact_archive_binary = base64.b64decode(model_data)
        archive_file = f"{model_tmp_location}.tar.gz"

        with open(f"{archive_file}", "wb") as f:
            f.write(model_artifact_archive_binary)

        with tarfile.open(f"{archive_file}") as archive:
            archive.extractall(model_tmp_location)

        loaded_model = tensorflow.keras.models.load_model(model_tmp_location)

        shutil.rmtree(model_tmp_location)
        os.remove(archive_file)

        return loaded_model, inference_method_name



from slai.loaders.torch_loader import TorchLoader
from slai.loaders.sklearn_loader import SklearnLoader
from slai.loaders.keras_loader import KerasLoader
from slai.exceptions import UnsupportedModelFormat


class ValidModelFrameworks:
    Torch = "TORCH"
    Sklearn = "SKLEARN"
    Keras = "KERAS"


class ModelLoader:
    @classmethod
    def load_model(cls, *, model_metadata, model_data, artifact_type):
        loaded_model = None

        if artifact_type == ValidModelFrameworks.Torch:
            loaded_model, inference_method_name = TorchLoader.load_model(
                model_metadata=model_metadata,
                model_data=model_data,
            )
        elif artifact_type == ValidModelFrameworks.Sklearn:
            loaded_model, inference_method_name = SklearnLoader.load_model(
                model_metadata=model_metadata,
                model_data=model_data,
            )
        elif artifact_type == ValidModelFrameworks.Keras:
            loaded_model, inference_method_name = KerasLoader.load_model(
                model_metadata=model_metadata,
                model_data=model_data,
            )
        else:
            raise UnsupportedModelFormat("invalid_artifact_type")

        return loaded_model, inference_method_name



import os
import xxhash
import shutil
import requests
import pathlib
import errno

from http import HTTPStatus
from tqdm import tqdm
from multiprocessing.pool import ThreadPool
from collections import namedtuple
from slai.clients.cli import get_cli_client
from slai.exceptions import DatasetNotFound
from slai.constants import DATASET_IO_THREAD_COUNT

DataSetNode = namedtuple("DataSetNode", "name parent_name hash index")


class DataSet:
    def __init__(
        self,
        model_version_id,
        local_path=None,
        version=None,
        create=False,
    ):
        self.model_version_id = model_version_id
        self.local_path = local_path
        self.create = create
        self.version = version
        self.metatable = {}

        self.cli_client = get_cli_client()

    def add_file(self, file_path, dir=None):
        if not self.local_path:
            raise RuntimeError("download_dataset_first")

        if dir is not None:
            shutil.copy(file_path, f"{self.local_path}/{dir}/")
        else:
            shutil.copy(file_path, f"{self.local_path}/")

    def add_files(self, files=[], dir=None):
        for f in files:
            self.add_file(f, dir)

    def add_directory(self, *, path, dir=None):
        if dir is not None:
            shutil.copytree(path, f"{self.local_path}/{dir}/{path}")
        else:
            shutil.copytree(path, f"{self.local_path}/{path}")

    def _handle_bad_request(self, response_body):
        msg = response_body["detail"]["non_field_errors"][0]
        if msg == "dataset_not_found":
            raise DatasetNotFound

    def download(self):
        try:
            self.dataset_object = self.cli_client.retrieve_dataset(
                model_version_id=self.model_version_id,
                version=self.version,
                create=self.create,
            )
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == HTTPStatus.BAD_REQUEST:
                self._handle_bad_request(e.response.json())
                return
            else:
                raise

        self.metatable = self._load_metatable()
        self.diff_table = {}

        if not self.local_path:
            self.local_path = f"./{self.name}"
            pathlib.Path(self.local_path).mkdir(parents=True, exist_ok=True)

        objects = []
        for file_id, obj in self.dataset_object["object_urls"].items():
            objects.append({"url": obj["url"], "path": obj["path"], "file_id": file_id})

        res_map = self._async_pool(objects, self._fetch_file)
        for file_id, success in res_map.items():
            if not success:
                raise RuntimeError("download_failed")

    def _fetch_file(self, obj):
        try:
            r = requests.get(obj["url"], stream=True)
        except requests.exceptions.ConnectionError:
            return (obj["file_id"], False)

        if r.status_code == HTTPStatus.OK:
            file = f"{self.local_path}/{obj['path']}"

            if not os.path.exists(os.path.dirname(file)):
                try:
                    os.makedirs(os.path.dirname(file))
                except OSError as exc:
                    if exc.errno != errno.EEXIST:
                        raise

            with open(f"{file}", "wb") as f:
                for chunk in r:
                    f.write(chunk)

        return (obj["file_id"], True)

    def _put_file(self, obj):
        with open(f"{self.local_path}/{obj['path']}", "rb") as f:
            try:
                r = requests.put(
                    f"{obj['url']}",
                    data=f.read(),
                )
            except requests.exceptions.ConnectionError:
                return (obj["file_id"], False)

            r.raise_for_status()

        return (obj["file_id"], True)

    def _async_pool(self, objects, callback):
        res_map = {}

        with tqdm(total=len(objects)) as progress_bar:
            for i, res in tqdm(
                enumerate(
                    ThreadPool(DATASET_IO_THREAD_COUNT).imap_unordered(
                        callback, objects
                    )
                )
            ):
                progress_bar.update()
                res_map[res[0]] = res[1]

        progress_bar.close()
        return res_map

    def path(self):
        return self.local_path

    def diff(self):
        self._checksum_local_path()

        for file_id, value in self.diff_table.items():
            print(f"{file_id}: {value}")

    def save(self):
        changes = self._checksum_local_path()
        if changes:
            self.dataset_object = self.cli_client.update_dataset(
                model_version_id=self.model_version_id,
                diff_table=self.diff_table,
            )

            objects = []
            for file_id, obj in self.dataset_object["object_urls"].items():
                objects.append(
                    {"url": obj["url"], "path": obj["path"], "file_id": file_id}
                )

            res_map = self._async_pool(objects, self._put_file)
            for file_id, success in res_map.items():
                if not success:
                    raise RuntimeError("upload_failed")

            self.metatable = self._load_metatable()
            self.diff_table = {}

    def describe(self):
        if not self.metatable:
            return

        return self.metatable.values()

    def reset(self):
        raise NotImplementedError

    def _load_metatable(self):
        metatable = {}
        for file_id, node in self.dataset_object["metatable"].items():
            node = DataSetNode(
                name=node[0], parent_name=node[1], hash=node[2], index=node[3]
            )
            metatable[file_id] = node

        return metatable

    def _regenerate_metatable(self):
        metatable = {}
        for dirpath, _, fnames in os.walk(self.local_path):
            path = dirpath.replace(str(self.local_path), "", 1)
            if len(path) > 0 and path[0] == "/":
                path = path[1:]

            for f in fnames:
                node = DataSetNode(
                    name=f,
                    parent_name=path,
                    hash=xxhash.xxh128(
                        open(f"{self.local_path}/{path}/{f}", "rb").read()
                    ).hexdigest(),
                    index=0,
                )
                file_id = xxhash.xxh128_hexdigest(f"{path}/{f}")
                metatable[file_id] = node

        return metatable

    def _checksum_local_path(self):
        self.diff_table = {}

        changes = False

        if not self.local_path:
            raise RuntimeError("download_dataset_first")

        self.next_metatable = self._regenerate_metatable()

        for file_id, node in self.next_metatable.items():
            if file_id in self.metatable:
                existing_node = self.metatable[file_id]

                if node.hash != existing_node.hash:
                    node = node._replace(index=existing_node.index + 1)
                    self.diff_table[file_id] = {
                        "node": node._asdict(),
                        "action": "modified",
                    }
                    changes = True
            else:
                node = node._replace(index=0)

                self.diff_table[file_id] = {
                    "node": node._asdict(),
                    "action": "created",
                }
                changes = True

        file_ids = set(self.metatable.keys())
        next_file_ids = set(self.next_metatable.keys())

        deleted_file_ids = file_ids.difference(next_file_ids)

        if len(deleted_file_ids) > 0:
            changes = True

        for file_id in list(deleted_file_ids):
            self.diff_table[file_id] = {
                "node": self.metatable[file_id]._asdict(),
                "action": "deleted",
            }

        return changes



import io
import base64
import inspect
import json
import pickle
import uuid
import tarfile
import shutil
import os

from importlib import import_module

TMP_MODEL_ARTIFACT_PATH_ROOT = "/tmp/artifacts"


class ModelSaver:
    @staticmethod
    def save_pytorch(model):
        _torch = import_module("torch")
        model_state_dict_binary = io.BytesIO()
        model_class_path = inspect.getfile(model.__class__).split("/")
        model_class_file_location = None

        if "site-packages" in model_class_path:
            index = model_class_path.index("site-packages")
            model_class_path = model_class_path[index + 1 :]
            model_class_file_location = ".".join(model_class_path)
        else:
            model_class_file_location = model_class_path

        _torch.save(model.state_dict(), model_state_dict_binary)

        # model parameters
        model_state_dict_binary_base64 = base64.b64encode(
            model_state_dict_binary.getvalue()
        ).decode()

        model_artifact = {
            "state_dict": model_state_dict_binary_base64,
        }

        model_metadata = {
            "model_class_file_location": model_class_file_location,
            "model_class_name": model.__class__.__name__,
        }

        model_artifact_base_64 = base64.b64encode(
            json.dumps(model_artifact).encode("utf-8")
        ).decode()

        return model_artifact_base_64, model_metadata

    @staticmethod
    def save_sklearn(model):
        model_binary = pickle.dumps(model)
        model_artifact_base64 = base64.b64encode(model_binary).decode()
        return model_artifact_base64

    @staticmethod
    def save_fastai(model):
        return None

    @staticmethod
    def save_keras(model):
        model_artifact_base64 = None
        model_tmp_location = f"{TMP_MODEL_ARTIFACT_PATH_ROOT}/{uuid.uuid4()}"

        try:
            model.save(model_tmp_location)

            with tarfile.open(f"{model_tmp_location}.tar.gz", "w:gz") as tar:
                tar.add(model_tmp_location, arcname="/")

            archive_file = f"{model_tmp_location}.tar.gz"
            with open(archive_file, "rb") as f:
                model_artifact_base64 = base64.b64encode(f.read()).decode()

            shutil.rmtree(model_tmp_location)
            os.remove(archive_file)
        except AttributeError:
            pass

        return model_artifact_base64, {}

    @staticmethod
    def save_tf(model):
        return None



class InvalidServicePrefix(Exception):
    pass


class ModelNotFound(Exception):
    pass


class InvalidRuntimeException(Exception):
    pass


class UnsupportedModelFormat(Exception):
    pass


class NoCredentialsFound(Exception):
    pass


class InvalidCredentials(Exception):
    pass


class IntegrationError(Exception):
    pass


class InvalidModelURI(Exception):
    pass


class DatasetNotFound(Exception):
    pass



import pickle
import base64

from slai.loaders import base_loader


class SklearnLoader(base_loader.BaseLoader):
    @classmethod
    def load_model(cls, model_metadata, model_data):
        model_artifact_model_binary = base64.b64decode(model_data)

        loaded_model_class = pickle.loads(model_artifact_model_binary)
        loaded_model = loaded_model_class()

        inference_method_name = "predict"
        return loaded_model, inference_method_name



import io
import base64

from importlib import import_module
from slai.loaders import base_loader


class TorchLoader(base_loader.BaseLoader):
    @classmethod
    def load_model(cls, model_metadata, model_data):
        model_framework_module = import_module("torch")

        model_artifact_model_state_dict = base64.b64decode(
            eval(model_data)["state_dict"]
        )
        model_file_module = ".".join(
            model_metadata["model_class_file_location"].split(".")[:-1]
        )
        model_class_name = model_metadata["model_class_name"]

        model_module = import_module(f"{model_file_module}")
        model_class = getattr(model_module, model_class_name)

        # instantiate model class
        loaded_model = model_class()

        # load trained model parameters
        model_state_dict_binary = io.BytesIO(model_artifact_model_state_dict)
        _state_dict = model_framework_module.load(model_state_dict_binary)
        loaded_model.load_state_dict(_state_dict)
        loaded_model.eval()

        return loaded_model, None
